(function() {
var GridViewColumnHelper = ASPx.CreateClass(null, {
 constructor: function(grid) {
  this.grid = grid;
  this.visibleHeaderTree = null;
  this.visibleDataRowTree = null;
  this.headerLayout = null;
  this.dataRowLayout = null;
  this.headerLeafs = null;
  this.dataRowLeafs = null;
  this.headerNodeHash = { };
  this.dataRowNodeHash = { };
  this.sortedVisibleChildColumnsHash = { };
  this.childColumnHash = { };
  this.fixedColumnsHash = null;
  this.visibleColumnIndices = null;
  this.rightTreeLines = null;
  this.cachedGroupColumnCount = -1;
 },
 GetAllColumns: function() { return this.grid.columns; },
 GetShowGroupedColumns: function() { return this.grid.showGroupedColumns; },
 FindColumn: function(index) {
  index = parseInt(index);
  return this.GetAllColumns().filter(function(c) { return c.index === index; })[0] || null;
 },
 HasColumnChildren: function(index) {
  return this.GetAllColumns().filter(function(c) { return c.parentIndex === index; }).length > 0;
 },
 GetGroupColumnCount: function() { 
  if(this.cachedGroupColumnCount === -1)
   this.cachedGroupColumnCount = this.GetAllColumns().filter(function(c) { return c.groupIndex > -1; }).length;
  return this.cachedGroupColumnCount;
 },
 GetParentColumnIndices: function(columnIndex) {
  var node = this.FindNode(columnIndex);
  if(!node)
   return [ ];
  return this.CreateTreeLine(node, false).map(function(n) { return n.columnIndex; });
 },
 FindColumnLayoutPos: function(layout, columnIndex) {
  for(var i = 0; i < layout.length; i++) {
   var level = layout[i];
   for(var j = 0; j < level.length; j++) {
    if(level[j].columnIndex === columnIndex)
     return { levelIndex: i, cellIndex: j };
   }
  }
  return { levelIndex: -1, cellIndex: -1 };
 },
 FindNode: function(columnIndex, isDataRowTree) {
  this.EnsureTrees();
  var hash = isDataRowTree ? this.dataRowNodeHash : this.headerNodeHash;
  return hash[columnIndex];
 },
 IsColumnFixed: function(columnIndex) {
  this.EnsureFixedColumnsHash();
  return this.fixedColumnsHash[columnIndex];
 },
 EnsureFixedColumnsHash: function() {
  if(this.fixedColumnsHash) 
   return;
  var firstLevelColumnsFixedState = { };
  var firstLevelColumns = this.GetVisibleChildColumns(-1, true);
  for(var i = 0; i < firstLevelColumns.length; i++) {
   var column = firstLevelColumns[i];
   firstLevelColumnsFixedState[column.index] = this.IsColumnFixedRecursive(column);
  }
  this.fixedColumnsHash = { };
  var allColumns = this.GetAllColumns();
  for(var i = 0; i < allColumns.length; i++) {
   var column = allColumns[i];
   var firstLevelColumn = column;
   while(firstLevelColumn.parentIndex >= 0)
    firstLevelColumn = this.FindColumn(firstLevelColumn.parentIndex);
   this.fixedColumnsHash[column.index] = firstLevelColumnsFixedState[firstLevelColumn.index];
  }
 },
 IsColumnFixedRecursive: function(column) {
  if(column.isFixed) return true;
  var childColumns = this.GetVisibleChildColumns(column.index, true);
  for(var i = 0; i < childColumns.length; i++) {
   if(this.IsColumnFixedRecursive(childColumns[i]))
    return true;
  }
  return false;
 },
 IsColumnInRightTreeLine: function(columnIndex, isHeader) {
  this.EnsureRightTreeLines();
  var treeLine = isHeader ? this.rightTreeLines.header : this.rightTreeLines.data;
  return treeLine.filter(function(n) { return n.columnIndex === columnIndex; }).length !== 0;
 },
 EnsureRightTreeLines: function() {
  if(this.rightTreeLines) return;
  this.rightTreeLines = { };
  var headerLeafs = this.GetHeaderLeafs();
  var dataRowLeafs = this.GetDataRowLeafs();
  this.rightTreeLines.header = this.CreateTreeLine(headerLeafs[headerLeafs.length - 1], true);
  this.rightTreeLines.data = this.CreateTreeLine(dataRowLeafs[dataRowLeafs.length - 1], true);
 },
 GetVisibleColumnIndices: function(isDataRowLayout) {
  if(!this.visibleColumnIndices)
   this.visibleColumnIndices = { header: this.CalculateVisibleColumnIndices(false), data: this.CalculateVisibleColumnIndices(true) };
  return isDataRowLayout ? this.visibleColumnIndices.data : this.visibleColumnIndices.header;
 },
 CalculateVisibleColumnIndices: function(isDataRowLayout) {
  var layout = isDataRowLayout ? this.GetDataRowLayout() : this.GetHeaderLayout();
  return ASPx.Data.ArrayFlatten(layout).map(function(n) { return n.columnIndex; });
 },
 GetAdaptiveRowInfo: function() {
  var adaptiveNodes = [ ];
  var tree = this.GetVisibleDataRowTree();
  this.FindAdaptiveNodes(tree, adaptiveNodes, false);
  for(var i = 0; i < adaptiveNodes.length; i++) {
   var node = adaptiveNodes[i];
   ASPx.Data.ArrayRemove(node.parent.children, node);
  }
  var dataRowLayout = [ ];
  this.PopulateRowLayout(tree, dataRowLayout);
  this.Invalidate();
  return { 
   dataRowLayout: dataRowLayout, 
   adaptiveColumnIndices: adaptiveNodes.map(function(n) { return n.columnIndex; }) 
  };
 },
 FindAdaptiveNodes: function(node, adaptiveNodes, isAdaptiveParent) {
  var isColumnAdaptive = node.column && node.column.adaptiveHidden;
  if(isColumnAdaptive || isAdaptiveParent) {
   adaptiveNodes.push(node);
   isAdaptiveParent |= isColumnAdaptive;
  }
  for(var i = 0; i < node.children.length; i++)
   this.FindAdaptiveNodes(node.children[i], adaptiveNodes, isAdaptiveParent);
 },
 EnsureTrees: function() {
  this.GetVisibleHeaderTree();
  this.GetVisibleDataRowTree();
 },
 GetVisibleHeaderTree: function() {
  if(!this.visibleHeaderTree) {
   var root = this.CreateTreeNode(-1, null);
   this.PopulateVisualTreeNode(root, this.GetVisibleChildColumns(-1), false);
   this.PopulateHeaderNodeSpans(root);
   this.visibleHeaderTree = root;
  }
  return this.visibleHeaderTree;
 },
 GetVisibleDataRowTree: function() {
  if(!this.visibleDataRowTree) {
   var root = this.CreateTreeNode(-1, null);
   this.PopulateVisualTreeNode(root, this.GetVisibleChildColumns(-1), true);
   this.PopulateDataRowNodeSpans(root);
   this.visibleDataRowTree = root;
  }
  return this.visibleDataRowTree;
 },
 GetHeaderLayout: function() {
  if(!this.headerLayout) {
   this.headerLayout = [ ];
   this.PopulateRowLayout(this.GetVisibleHeaderTree(), this.headerLayout);
  }
  return this.headerLayout;
 },
 GetDataRowLayout: function() {
  if(!this.dataRowLayout) {
   this.dataRowLayout = [ ];
   this.PopulateRowLayout(this.GetVisibleDataRowTree(), this.dataRowLayout);
  }
  return this.dataRowLayout;
 },
 GetHeaderLeafs: function() { 
  if(!this.headerLeafs) {
   this.headerLeafs = [ ];
   this.PopulateLeafs(this.GetVisibleHeaderTree(), this.headerLeafs);
  }
  return this.headerLeafs;
 },
 GetDataRowLeafs: function() { 
  if(!this.dataRowLeafs) {
   this.dataRowLeafs = [ ];
   this.PopulateLeafs(this.GetVisibleDataRowTree(), this.dataRowLeafs);
  }
  return this.dataRowLeafs;
 },
 PopulateVisualTreeNode: function(node, visibleColumns, isDataRowTree) {
  var nodeHash = isDataRowTree ? this.dataRowNodeHash : this.headerNodeHash;
  for(var i = 0; i < visibleColumns.length; i++) {
   var bandNode = node;
   var column = visibleColumns[i];
   var childColumns = this.GetVisibleChildColumns(column.index);
   if(this.UseColumnInVisualTree(column, childColumns, isDataRowTree)) {
    var child = this.CreateTreeNode(column.index, node);
    node.children.push(child);
    nodeHash[column.index] = child;
    bandNode = child;
   }
   if(childColumns.length > 0)
    this.PopulateVisualTreeNode(bandNode, childColumns, isDataRowTree);
  }
 },
 UseColumnInVisualTree: function(column, childColumns, isDataRowTree) {
  if(column.groupIndex > -1 && !this.GetShowGroupedColumns())
   return false;
  if(isDataRowTree && column.isBand) {
   var childBandColumns = childColumns.filter(function(c) { return c.isBand; });
   if(childBandColumns.length > 0)
    return false;
   var visibleChildrenInTree = childColumns.filter(function(c) { return this.UseColumnInVisualTree(c, this.GetVisibleChildColumns(c.index), isDataRowTree); }.bind(this));
   return visibleChildrenInTree.length === 0;
  }
  return true;
 },
 GetChildColumns: function(columnIndex) {
  if(!this.childColumnHash[columnIndex])
   this.childColumnHash[columnIndex] = this.GetAllColumns().filter(function(c) { return c.parentIndex === columnIndex; });
  return this.childColumnHash[columnIndex];
 },
 GetVisibleChildColumns: function(columnIndex, unsorted) {
  columnIndex = parseInt(columnIndex);
  if(unsorted)
   return this.FindVisibleChildColumns(columnIndex, unsorted);
  if(!this.sortedVisibleChildColumnsHash[columnIndex])
   this.sortedVisibleChildColumnsHash[columnIndex] = this.FindVisibleChildColumns(columnIndex);
  return this.sortedVisibleChildColumnsHash[columnIndex];
 },
 FindVisibleChildColumns: function(columnIndex, unsorted) {
  var columns = this.GetChildColumns(columnIndex).filter(function(c) { return c.visible; }.bind(this));
  if(!unsorted)
   columns.sort(function(c1, c2) {
    var isRootLevel = columnIndex === -1;
    if(isRootLevel && this.IsColumnFixed(c1.index) !== this.IsColumnFixed(c2.index))
     return this.IsColumnFixed(c1.index) ? -1 : 1;
    if(c1.visibleIndex > c2.visibleIndex)
     return 1;
    if(c1.visibleIndex < c2.visibleIndex)
     return -1;
    return c1.collectionIndex - c2.collectionIndex;
   }.aspxBind(this));
  return columns;
 },
 PopulateRowLayout: function(node, layout) {
  for(var i = 0; i < node.children.length; i++) {
   var child = node.children[i];
   var levelIndex = this.FindNodeLayoutLevelIndex(child);
   while(layout.length < levelIndex + 1)
    layout.push([ ]);
   layout[levelIndex].push(child);
   this.PopulateRowLayout(child, layout);
  }
 },
 PopulateLeafs: function(node, leafs) {
  for(var i = 0; i < node.children.length; i++) {
   var child = node.children[i];
   if(child.children.length > 0)
    this.PopulateLeafs(child, leafs);
   else
    leafs.push(child);
  }
 },
 FindNodeLayoutLevelIndex: function(node) {
  var level = 0;
  node = node.parent;
  while(node.parent) {
   level += node.rowSpan;
   node = node.parent;
  }
  return level;
 },
 PopulateDataRowNodeSpans: function(root) {
  this.GetVisibleHeaderTree(); 
  root.rowSpan = root.colSpan = 1;
  var useBandedRowLayout = true; 
  for(var key in this.dataRowNodeHash) {
   if(!this.dataRowNodeHash.hasOwnProperty(key)) continue;
   var dataRowNode = this.dataRowNodeHash[key];
   var headerNode = this.headerNodeHash[key];
   dataRowNode.colSpan = headerNode.colSpan;
   dataRowNode.rowSpan = useBandedRowLayout ? headerNode.rowSpan : 1;
  }
  var bandColumnIndices = [ ];
  for(var key in this.headerNodeHash) {
   if(!this.headerNodeHash.hasOwnProperty(key)) continue;
   var headerNode = this.headerNodeHash[key];
   if(headerNode.column && headerNode.column.isBand)
    bandColumnIndices.push(headerNode.column.index);
  }
  if(bandColumnIndices.length > 0 && useBandedRowLayout)
   this.NormalizeDataRowTreeNodeRowSpans(bandColumnIndices);
 },
 NormalizeDataRowTreeNodeRowSpans: function(bandColumnIndices) {
  var matrix = this.CreateLayoutMatrixInfo(this.GetHeaderLayout()).matrix;
  var height = matrix.length;
  var width = matrix[0].length;
  for(var j = 0; j < width; j++) {
   var lastNotBandColumnIndex = matrix[height - 1][j];
   for(var i = height - 2; i >= 0; i--) {
    var columnIndex = matrix[i][j];
    if(bandColumnIndices.indexOf(columnIndex) > -1)
     matrix[i][j] = lastNotBandColumnIndex;
    else 
     lastNotBandColumnIndex = columnIndex;
   }
  }
  for(var i = height - 1; i > 0; i--) {
   if(this.IsMatrixRowEqual(matrix, i, i - 1))
    matrix.splice(i, 1);
  }
  height = matrix.length;
  var columnRowSpan = { };
  for(var j = 0; j < width; j++) {
   var columnIndices = this.GetColumnIndicesByColInd(height, j, matrix);
   var groupedIndices = columnIndices.reduce(function(r, cI) { (r[cI] = r[cI] || [ ]).push(cI); return r; }, { });
   for(var columnIndex in groupedIndices) {
    if(!groupedIndices.hasOwnProperty(columnIndex)) continue;
    columnRowSpan[columnIndex] = groupedIndices[columnIndex].length;
   }
  }
  for(var columnIndex in columnRowSpan) {
   if(!columnRowSpan.hasOwnProperty(columnIndex)) continue;
   this.dataRowNodeHash[columnIndex].rowSpan = columnRowSpan[columnIndex];
  }
 },
 GetColumnIndicesByColInd: function(height, colInd, matrix) {
  return ASPx.Data.Range(height).map(function(i) { return matrix[i][colInd]; });
 },
 PopulateHeaderNodeSpans: function(root) {
  var leafs = [ ];
  this.PopulateLeafs(root, leafs);
  if(leafs.length === 0)
   return;
  var leafTreeLines = [ ];
  for(var i = 0; i < leafs.length; i++) {
   var treeLine = [ ];
   var current = leafs[i];
   while(current != null) {
    current.colSpan |= 0;
    current.colSpan++;
    current.rowSpan = current.column ? current.column.rowSpan : 1;
    treeLine.push(current);
    current = current.parent;
   }
   leafTreeLines.push(treeLine);
  }
  var normalizeIterationCount = 2;
  for(var i = 0; i < normalizeIterationCount; i++)
   this.NormalizeHeaderNodesRowSpan(leafTreeLines);
 },
 NormalizeHeaderNodesRowSpan: function(leafTreeLines) {
  var height = leafTreeLines.reduce(function(prevValue, treeLine) { return Math.max(prevValue, treeLine.length); }, 0);
  for(var i = 0; i < leafTreeLines.length; i++) {
   var treeLine = leafTreeLines[i];
   var rowSpanHeight = treeLine.reduce(function(prevValue, node) { return prevValue + node.rowSpan; }, 0);
   var diff = height - rowSpanHeight;
   if(diff < 0)
    this.CalculateTreeLineRowSpanProportionally(treeLine, height);
   else
    treeLine[0].rowSpan += diff; 
  }
 },
 CalculateTreeLineRowSpanProportionally: function(treeLine, requiredHeight) {
  var rowSpans = this.CalculateTreeLineRowSpanProportionallyCore(treeLine, requiredHeight);
  for(var i = 0; i < treeLine.length; i++)
   treeLine[i].rowSpan = rowSpans[i];
 },
 CalculateTreeLineRowSpanProportionallyCore: function(treeLine, requiredHeight) {
  var result = [ ];
  var height = treeLine.reduce(function(prevValue, node) { return prevValue + node.rowSpan; }, 0);
  var delta = requiredHeight - height;
  if(delta >= 0) 
   return result;
  delta *= -1;
  var excess = delta;
  for(var i = 0; i < treeLine.length; i++) {
   var node = treeLine[i];
   result[i] = node.rowSpan;
   if(excess <= 0) continue;
   var k = Math.ceil(delta * node.rowSpan / height) + 1;
   if(k >= node.rowSpan)
    k = node.rowSpan - 1;
   excess -= k;
   if(excess < 0)
    k += excess;
   result[i] -= k;
  }
  return result;
 },
 CreateTreeNode: function(columnIndex, parent) {
  if(parent === undefined)
   parent = null;
  return { columnIndex: columnIndex, children: [ ], parent: parent, column: this.FindColumn(columnIndex) };
 },
 CreateTreeLine: function(node, insertCurrent) {
  var result = [ ];
  var current = insertCurrent ? node : node.parent;
  while(current.column !== null) {
   result.push(current);
   current = current.parent;
  }
  return result;
 },
 Invalidate: function() {
  this.visibleHeaderTree = null;
  this.visibleDataRowTree = null;
  this.headerLayout = null;
  this.dataRowLayout = null;
  this.headerLeafs = null;
  this.dataRowLeafs = null;
  this.headerNodeHash = { };
  this.dataRowNodeHash = { };
  this.sortedVisibleChildColumnsHash = { };
  this.childColumnHash = { };
  this.fixedColumnsHash = null;
  this.visibleColumnIndices = null;
  this.rightTreeLines = null;
  this.cachedGroupColumnCount = -1;
 },
 GetOrderedColumns: function() {
  var columns = this.GetAllColumns();
  columns.sort(function(c1, c2) {
   if(c1.index > c2.index)
    return 1;
   if(c1.index < c2.index)
    return -1;
   return 0;
  });
  return columns;
 },
 SerializeColumnVisibleState: function() {
  var state = [ ];
  var columns = this.GetOrderedColumns();
  state.push("clientvis" + columns.length);
  for(var i = 0; i < columns.length; i++) {
   var column = columns[i];
   state.push((column.visible ? "t" : "f") + column.visibleIndex);
  }
  return state.join("|");
 },
 SerializeColumnHierarchyState: function() {
  var state = [ ];
  var columns = this.GetOrderedColumns();
  state.push("clienthierar" + columns.length);
  for(var i = 0; i < columns.length; i++) {
   var column = columns[i];
   state.push(column.index + "|" + column.parentIndex);
  }
  return state.join("|");
 },
 CreateLayoutMatrixInfo: function(layout) {
  var matrix = [ ];
  var inverseMatrix = { };
  for(var i = 0; i < layout.length; i++) {
   var lastFreeIndex = 0;
   var level = layout[i];
   for(var j = 0; j < level.length; j++) {
    var node = level[j];
    lastFreeIndex = this.FindMatrixFreeCellIndex(matrix, i, lastFreeIndex);
    this.FillMatrixBlock(matrix, inverseMatrix, i, lastFreeIndex, node.rowSpan, node.colSpan, node.columnIndex);
    lastFreeIndex += node.colSpan;
   }
  }
  return { matrix: matrix, inverseMatrix: inverseMatrix };
 },
 FindMatrixFreeCellIndex: function(matrix, rowIndex, lastFreeCell) {
  var row = matrix[rowIndex];
  var result = lastFreeCell;
  if(row) {
   while(!isNaN(row[result]))
    result++;
  } 
  return result;
 },
 FillMatrixBlock: function(matrix, inverseMatrix, rowIndex, cellIndex, rowSpan, colSpan, columnIndex) {
  var rect = {
   top: rowIndex,
   bottom: rowIndex + rowSpan - 1,
   left: cellIndex,
   right: cellIndex + colSpan - 1
  };
  for(var i = rect.top; i <= rect.bottom; i++) {
   while(!matrix[i])
    matrix.push([]);
   for(var j = rect.left; j <= rect.right; j++)
    matrix[i][j] = columnIndex;
  }
  inverseMatrix[columnIndex] = rect;
 },
 IsMatrixRowEqual: function(matrix, i1, i2) {
  var row1 = matrix[i1];
  var row2 = matrix[i2];
  for(var j = 0; j < row1.length; j++) {
   if(row1[j] !== row2[j])
    return false;
  }
  return true;
 }
});
ASPx.GridViewColumnHelper = GridViewColumnHelper;
})();
